from playa.data._asobj import asobj
from playa.data.content import (
    Color,
    ColorSpace,
    DashPattern,
    Glyph,
    GraphicState,
    Image,
    MarkedContent,
    Path,
    PathSegment,
    Tag,
    Text,
)
from playa.data.metadata import (
    Annotation,
    Destination,
    Document,
    Encryption,
    Font,
    IndirectObject,
    Outline,
    Page,
    Resources,
    StreamObject,
    StructElement,
    StructTree,
    XObject,
)

__all__ = [
    "Color",
    "ColorSpace",
    "DashPattern",
    "Glyph",
    "GraphicState",
    "Image",
    "MarkedContent",
    "Path",
    "PathSegment",
    "Tag",
    "Text",
    "Annotation",
    "Destination",
    "Document",
    "Encryption",
    "Font",
    "IndirectObject",
    "Outline",
    "Page",
    "Resources",
    "StreamObject",
    "StructElement",
    "StructTree",
    "XObject",
    "asobj",
]
VERSION = "1.0"
